#
# This file generates the sidebar/toctree for all TileDB projects and should
# be copied to each project when it is updated.
#
# This file is originally from the RobotPy documentation project
# https://github.com/robotpy/robotpy-docs, licensed under Apache v2.
#

import os

def write_if_changed(fname, contents):

    try:
        with open(fname, 'r') as fp:
            old_contents = fp.read()
    except:
        old_contents = ''

    if old_contents != contents:
        with open(fname, 'w') as fp:
            fp.write(contents)

def generate_sidebar(conf, conf_api):

    version = conf['rtd_version']

    lines = [
        '', '.. DO NOT MODIFY! THIS PAGE IS AUTOGENERATED!',
        '   To edit the sidebar, modify gensidebar.py and re-build the docs.', ''
    ]

    url_base = 'https://docs.tiledb.io'
    lang = 'en'

    def toctree(name):
        lines.extend(['.. toctree::',
                      '    :caption: %s' % name,
                      '    :maxdepth: 1',
                      ''])

    def endl():
        lines.append('')

    def write(desc, link):
        if conf_api == 'tiledb':
            args = desc, link
        else:
            args = desc, '%s/%s/%s/%s.html' % (url_base, lang, version, link)

        lines.append('    %s <%s>' % args)

    def write_api(project, desc, rst_page):
        # From non-root project to root project link
        if project == 'tiledb' and conf_api != 'tiledb':
            args = desc, url_base, lang, version, rst_page
            lines.append('    %s API <%s/%s/%s/%s.html>' % args)
        # From anything to non-root project link
        elif project != conf_api:
            args = desc, url_base, project, lang, version, rst_page
            lines.append('    %s API <%s/projects/%s/%s/%s/%s.html>' % args)
        # Local project link
        else:
            args = desc, rst_page
            lines.append('    %s API <%s>' % args)

    def write_api_url(desc, url):
        lines.append('    %s API <%s>' % (desc, url))

    #
    # Specify the sidebar contents here
    #

    toctree('Getting Started')
    write('Introduction', 'introduction')
    write('Quickstart', 'quickstart')
    write('Installation', 'installation')
    write('Usage', 'usage')
    endl()

    toctree('API References')
    write_api('tiledb', 'C', 'c-api')
    write_api('tiledb', 'C++', 'c++-api')
    write_api('tiledb-py', 'Python', 'python-api')
    write_api_url('R', 'https://tiledb-inc.github.io/TileDB-R/reference/index.html')
    write_api_url('Go', 'https://godoc.org/github.com/TileDB-Inc/TileDB-Go')
    endl()

    toctree('Tutorials - Beginner')
    write('Dense Arrays', 'tutorials/dense-arrays')
    write('Sparse Arrays', 'tutorials/sparse-arrays')
    write('Multi-attribute Arrays', 'tutorials/multi-attribute-arrays')
    write('Variable-length Attributes', 'tutorials/variable-length-attributes')
    write('Filters', 'tutorials/filters')
    write('Compression', 'tutorials/compression')
    write('Encryption', 'tutorials/encryption')
    write('Catching Errors', 'tutorials/errors')
    write('Working with S3', 'tutorials/working-with-s3')
    write('Working with HDFS', 'tutorials/working-with-hdfs')
    endl()

    toctree('Tutorials - Intermediate')
    write('Tiling Dense Arrays', 'tutorials/tiling-dense')
    write('Tiling Sparse Arrays', 'tutorials/tiling-sparse')
    write('Writing Sparse Arrays', 'tutorials/writing-sparse')
    write('Writing Dense Arrays', 'tutorials/writing-dense')
    write('Reading Arrays', 'tutorials/reading')
    write('Negative and Real Domains', 'tutorials/neg-real')
    write('Datetimes', 'tutorials/datetimes')
    write('Key-value Store', 'tutorials/kv')
    write('Object Management', 'tutorials/object')
    write('Virtual Filesystem', 'tutorials/vfs')
    write('Configuration Parameters', 'tutorials/config')
    endl()

    toctree('Tutorials - Advanced')
    write('Fragments and Consolidation', 'tutorials/fragments-consolidation')
    write('Advanced Consolidation', 'tutorials/advanced-consolidation')
    write('Asynchronous Queries', 'tutorials/async')
    write('Parallelism', 'tutorials/parallelism')
    write('Concurrency and Consistency', 'tutorials/concurrency-consistency')
    write('Format description', 'tutorials/format-description')
    endl()

    toctree('Performance')
    write('Introduction to Performance', 'performance/introduction')
    write('Using TileDB statistics', 'performance/using-tiledb-statistics')
    write('Performance factors', 'performance/performance-factors')
    endl()

    toctree('Real-world Examples')
    write('Dense Image Data', 'real-world-examples/dense-image-data')
    write('Sparse Geospatial Data', 'real-world-examples/sparse-geospatial-data')
    endl()

    write_if_changed('_sidebar.rst.inc', '\n'.join(lines))
